package com.jay.monitor.data.server.store;

import com.jay.monitor.data.core.enums.MsgTypeEnums;
import com.jay.monitor.data.server.config.props.MonitorProperties;
import com.jay.monitor.data.server.enums.DataSplitType;
import com.jay.monitor.data.server.enums.StoreType;
import com.jay.monitor.data.server.store.mysql.MysqlDDLProcess;
import com.jay.monitor.data.server.utils.MonitorPropertiesUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.atomic.LongAdder;

/**
 * @Module 执行器
 * @Description 抽象存储
 * @Author liukaixiong
 * @Date 2020/10/29 13:33
 */
public abstract class AbstractStoreProcess<T> implements Runnable, InitializingBean {
    private Logger logger = LoggerFactory.getLogger(getClass());

    private LongAdder counter = new LongAdder();

    private int queueSize = 5000;

    private int batchSize = 500;

    private BlockingQueue<T> queue = new ArrayBlockingQueue<>(queueSize);

    @Autowired
    private MysqlDDLProcess ddlProcess;

    @Autowired
    private MonitorProperties monitorProperties;

    public void setQueueSize(int queueSize) {
        this.queueSize = queueSize;
    }

    public int getQueueSize() {
        return queueSize;
    }

    public int getBatchSize() {
        return batchSize;
    }

    public void setBatchSize(int batchSize) {
        this.batchSize = batchSize;
    }

    /**
     * 存储类型
     *
     * @return
     */
    public abstract StoreType storeType();

    /**
     * 消息类型
     *
     * @return
     */
    public abstract MsgTypeEnums msgTypeEnums();

    public Object store(T request) throws Exception {

        storeBefore(request);

        boolean add = getQueue().offer(request);

        if (!add) {

            counter.decrement();

            if (counter.intValue() == 0 || counter.intValue() > queueSize) {
                logger.warn(getClass() + " 队列消费满了，装不下去了");
                counter.reset();
            }
        }

        storeAfter(request, add);

        return add;
    }

    /**
     * 存储之前触发
     *
     * @param request
     */
    protected void storeBefore(T request) {

    }

    /**
     * 存储之后触发
     *
     * @param request
     * @param success
     */
    protected void storeAfter(T request, boolean success) {
    }


    public abstract String searchContentById(String date, String id) throws Exception;

    public abstract void speedUpConsumerList(List<T> request);

    public void speedUpConsumer(BlockingQueue<T> queue, int rate) {
        List<T> requestList = new ArrayList<>();

        for (int i = 0; i < rate; i++) {
            T request = queue.poll();
            requestList.add(request);
        }

        speedUpConsumerList(requestList);
    }

    protected BlockingQueue<T> getQueue() {
        return queue;
    }

    @Override
    public void run() {
        logger.info("开启异步消费机...");
        int defaultSize = getQueueSize();
        int batchSize = getBatchSize();
        int minBatch = batchSize / 10;
        int thresholdValue = (int) (defaultSize * 0.8);
        while (true) {
            try {
                int size = this.queue.size();
                if (size > 0) {
                    if (size > thresholdValue) {
                        speedUpConsumer(this.queue, defaultSize / 2);
                    } else if (size > batchSize) {
                        speedUpConsumer(this.queue, batchSize);
                    } else if (size > minBatch) {
                        speedUpConsumer(this.queue, size);
                    } else {
                        // 如果数量不多的情况下
                        speedUpConsumer(this.queue, 1);
                    }
                }
            } catch (Exception e) {
                logger.warn("异步队列消费失败", e);
                sleep(5);
            } finally {
                sleep(5);
            }
        }
    }

    private void sleep(int ms) {
        try {
            Thread.sleep(ms);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    public void startConsumerThread() {
        String threadName = getClass().getSimpleName() + "-consumer-thread";
        logger.info("开启消费线程:" + threadName);
        new Thread(this, threadName).start();
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        startConsumerThread();
    }

    /**
     * 获取分区编号
     *
     * @return
     */
    protected String getPartitionId() {
        // 从配置文件中获取对应的规则
        Map<String, DataSplitType> tableRuleMap = MonitorPropertiesUtil.getMysqlTableRuleMap(monitorProperties);

        if (tableRuleMap != null) {
            DataSplitType dataSplitType = tableRuleMap.get(tableName());
            if (dataSplitType != null) {
                String dataId = ddlProcess.getDataId(tableName(), dataSplitType);
                return dataId;
            }
        }
        return null;
    }

    public abstract String tableName();
}
